/*
    mandelbmp - render the Mandelbrot set or Julia sets as BMP files
    Copyright (C) 2022 cas <cas@sdf.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "bmp.h"
#include "colormaps.h"
#include "cplx.h"

int main(int argc, char* argv[])
{
    if(argc < 11){
        printf("usage: %s Re{c} Im{c} nx ny x1 y1 x2 y2 max_iter output.bmp [colormap]\n", argv[0]);
        
        printf("good values for x1, y1, x2, y2 are -2, -2, 2 and 2\n\n");
        
        printf("colormaps:\n");
        printf(" 0 - nausea (default)\n");
        printf(" 1 - colorblind\n");
        printf(" 2 - inverted colorblind\n");
        printf(" 3 - heresy\n");
        printf(" 4 - prophecy\n");
        printf(" 5 - parasite\n");
        printf(" 6 - invader\n");
        exit(0);
    }
    
    double cr   = 0.;
    double ci   = 0.;
    
    int mode  = 1; // 0 is mandelbrot mode, 1 is julia mode
    if(argv[1][0] == 'x' && argv[2][0] == 'y'){
        mode = 0;
    }else{
        mode = 1;
        cr      = atof(argv[1]);
        ci      = atof(argv[2]);
    }
    
    int nx      = atoi(argv[3]); // resolution in x dir.
    int ny      = atoi(argv[4]); // resolution in x dir.
    double x1   = atof(argv[5]); // lowest real part
    double y1   = atof(argv[6]); // lowest imag. part
    double x2   = atof(argv[7]); // highest real part
    double y2   = atof(argv[8]); // highest imag. part
    int maxiter = atoi(argv[9]); // max. num. of iterations
    
    int colormap = 0; // default colormap is strega
    
    if(argc >= 12){
      colormap = atoi(argv[11]); // fetch color
    }
    
    // make space for a BMP image
    BMP* bmp = create_BMP(nx, ny);
    
    // make space for a complex number
    cplx z, z0;
    
    for(int i = 0; i < nx; ++i){
        double x = x1 + (x2 - x1) * ((double) i / nx);
        for(int j = 0; j < ny; ++j){
            double y = y1 + (y2 - y1) * ((double) j / ny);
            
            z.real = x;
            z.imag = y;
            
            if(mode == 0){
                z0.real = x; // Mandelbrot mode
                z0.imag = y;
            }else{
                z0.real = cr; // Julia mode
                z0.imag = ci;
            }
            
            double zabs2 = (z.real * z.real + z.imag * z.imag);
            
            int iter;
            for(iter = 0; iter < maxiter; ++iter){
                if(zabs2 > 4.)
                    break;

                cplx_power(&z, 2);
                cplx_add(&z, &z0);
                zabs2 = (z.real * z.real + z.imag * z.imag);
            }
            
            double r = (double) iter / maxiter;
            int c1;
            int c2;
            int c3;
            cmap_get_rgb(colormap, r, &c1, &c2, &c3);
            
            set_rgb(i, j, c1, c2, c3, bmp);
        }
    }
    
    write_BMP(argv[10], bmp);
    free_BMP(bmp);
}
